"""
The implementation file for all entities needed for ANAC-ANL 2024

Participants need to provide a class inherited from `SAONegotiator` that implements its __call__ method

"""
from .runner import *

__all__ = runner.__all__
